"""Define base module class."""
import torch
from torch import nn

from collections import defaultdict
from typing import Any, Optional, Dict, List, Union, Callable


class IterationModule(nn.Module):
    """A module where the forward pass is called iteratively.

    The `forward()` method calls `self.iteration()` iteratively, until either a maximum
    number of steps is reached, or `self.converged()` is true. Any arguments passed to
    `forward()` are passed along:

        self.iteration(*args, **kwargs)

    When done iterating, `forward()` returns the output from `self.post_iteration()`.
    The methods `self.pre_iteration()` and `self.post_iteration()` can also be used to
    perform any necessary pre- and post-processing, as they are called before the first
    iteration and after the last, respectively. They are passed the arguments passed to
    `forward()`:

        self.pre_iteration(*args, **kwargs)
        self.post_iteration(*args, **kwargs)

    By default, these do nothing and return nothing.
    """

    def __init__(self, max_iterations: int = 1000, **kwargs):
        """Initialize the module.

        :param max_iterations: maximum number of `iteration()` calls in one forward pass
        """
        super().__init__(**kwargs)
        self.max_iterations = max_iterations
        self.hooks = defaultdict(list)
        self.iteration_idx = -1

    def register_iteration_hook(self, kind: str, hook: Callable):
        """Register a hook for the forward iteration.

        Hooks can be assigned for the following events:
        * pre:          called before `pre_iteration()`
            signature: `hook(module)`
        * post:         called after `post_iteration()`
            signature: `hook(module)`
        * iteration:    called after every call to `iteration()` (before `converged()`)
            signature: `hook(module) -> bool`
            A truthful return value ends the iteration. The iteration index is available
            in `module.iteration_idx`.

        Multiple hooks can be attached to the same event.

        :param kind: the kind of hook to register
        :param hook: the function to be called
        """
        if kind not in ["pre", "post", "iteration"]:
            raise ValueError(f"Unknown event type in register_iteration_hook: {kind}")

        self.hooks[kind].append(hook)

    def forward(self, *args, **kwargs) -> Any:
        self.iteration_idx = -1

        self._call_hooks(self.hooks["pre"])
        self.pre_iteration(*args, **kwargs)
        for i in range(self.max_iterations):
            self.iteration_idx = i
            self.iteration(*args, **kwargs)
            if self._call_hooks(self.hooks["iteration"], check_output=True):
                break
            if self.converged(*args, **kwargs):
                break

        retval = self.post_iteration(*args, **kwargs)
        self._call_hooks(self.hooks["post"])

        return retval

    def iteration(self, *args, **kwargs):
        raise NotImplementedError(
            f'Module [{type(self).__name__}] is missing the required "iteration" function'
        )

    def converged(self, *args, **kwargs) -> bool:
        return False

    def pre_iteration(self, *args, **kwargs):
        pass

    def post_iteration(self, *args, **kwargs) -> Any:
        pass

    def _call_hooks(self, hooks: List[Callable], check_output: bool = False) -> bool:
        for hook in hooks:
            retval = hook(self)
            if check_output and retval:
                return True

        return False


class IterationLossModule(IterationModule):
    """A specialization of `IterationModule` where the iteration is derived from a loss
    function.

    This creates an optimizer in the `pre_iteration()`, then for each iteration runs
    `backward()` on the output from `self.iteration_loss()` and steps the optimizer.
    This is followed by an optional projection step; see `iteration_projection` below.
    Note that projection is a very simple way of enforcing constraints, and might not
    work well with adaptive step optimizers.

    The constructor has options for choosing the optimizer to use, as well as for an
    optional learning-rate scheduler; see below.

    Functions to implement:
      * `iteration_loss(*args, **kwargs)` should return the loss; the output is stored
        in `self.last_iteration_loss` as a number (i.e. `item()` is called on the tensor
        output from `iteration_loss()`)
      * `iteration_parameters()` should return a list of parameters to be optimized
        during the iteration

    Note that typically the `iteration_parameters()` should *not* be included in the
    module's `parameters()`, but should potentially be saved as part of the
    `state_dict`, so it is recommended that they be registered as buffers.

    The optimization and scheduling features of the class can be used in combination
    with manually calculated gradients by overriding

        iteration_set_gradients(*args, **kwargs)
    """

    def __init__(
        self,
        iteration_optimizer: Callable = torch.optim.SGD,  # type: ignore
        iteration_scheduler: Optional[Callable] = None,
        iteration_lr: float = 1.0,
        it_optim_kwargs: Optional[Dict[str, Any]] = None,
        it_sched_kwargs: Optional[Dict[str, Any]] = None,
        iteration_projection: Optional[Callable] = None,
        **kwargs,
    ) -> None:
        """Initialize the module.

        :param iteration_optimizer: optimizer to use for forward iteration; e.g.,
            `torch.optim.SGD`
        :param iteration_scheduler: scheduler to use (if any) for forward iteration;
            e.g., `torch.optim.lr_scheduler.StepLR`
        :param iteration_lr: learning rate for forward iteration; this is a shortcut
            that overrides any potential learning rate from `it_optim_kwargs`
        :param it_optim_kwargs: dictionary of keyword arguments to pass to the optimizer
        :param it_sched_kwargs: dictionary of keyword arguments to pass to the scheduler
        :param iteration_projection: optional projection to perform after each optimizer
            step; this should be a callable that will be applied to each element of
            `self.iteration_parameters()` (e.g., `torch.nn.functional.relu`)
        :param kwargs: other keyword arguments are passed to `IterationModule`
        """
        super().__init__(**kwargs)

        self.it_construct_optim = iteration_optimizer
        self.it_construct_sched = iteration_scheduler

        self.it_optim_kwargs = it_optim_kwargs if it_optim_kwargs is not None else {}
        self.it_optim_kwargs.setdefault("lr", iteration_lr)

        self.it_sched_kwargs = it_sched_kwargs if it_sched_kwargs is not None else {}

        self.iteration_optimizer: torch.optim.Optimizer = None  # type: ignore
        self.iteration_scheduler: Optional[torch.optim.lr_scheduler.LRScheduler] = None

        self.iteration_projection = iteration_projection

        self.last_iteration_loss = None

    def iteration(self, *args, **kwargs):
        self.iteration_set_gradients(*args, **kwargs)

        self.iteration_optimizer.step()
        if self.iteration_scheduler is not None:
            self.iteration_scheduler.step()

        if self.iteration_projection is not None:
            with torch.no_grad():
                for param in self.iteration_parameters():
                    param.data = self.iteration_projection(param.data)

    def pre_iteration(self, *args, **kwargs):
        """Pre-iteration processing.

        This sets `requires_grad` to `True` for the `iteration_parameters()` and to
        `False` for the `parameters()`. It also generates an optimizer and a scheduler,
        if one is requested.
        """
        super().pre_iteration(*args, **kwargs)

        # iteration parameters require grad...
        for param in self.iteration_parameters():
            param.requires_grad_(True)
        # ...but other parameters don't
        grad_state = []
        for param in self.parameters():
            grad_state.append(param.requires_grad)
            param.requires_grad_(False)
        self._pre_requires_grad_state = grad_state

        self.iteration_optimizer = self.it_construct_optim(
            self.iteration_parameters(), **self.it_optim_kwargs
        )

        if self.it_construct_sched is not None:
            self.iteration_scheduler = self.it_construct_sched(
                self.iteration_optimizer, **self.it_sched_kwargs
            )

    def post_iteration(self, *args, **kwargs) -> Any:
        """Post-iteration processing.

        This sets `requires_grad` to `False` for the `iteration_parameters()` and to
        whatever it was before the iteration for the `parameters()`.
        """
        # reset: no grad for iteration parameters...
        for param in self.iteration_parameters():
            param.requires_grad_(False)
        # ...and reset grad for the others
        for param, old_state in zip(self.parameters(), self._pre_requires_grad_state):
            param.requires_grad_(old_state)

        return super().post_iteration(*args, **kwargs)

    def iteration_set_gradients(self, *args, **kwargs):
        self.iteration_optimizer.zero_grad()

        loss = self.iteration_loss(*args, **kwargs)
        loss.backward()

        self.last_iteration_loss = loss.item()

    def iteration_loss(self, *args, **kwargs):
        raise NotImplementedError(
            f"Module [{type(self).__name__}] is missing the required "
            f'"iteration_loss" function'
        )

    def iteration_parameters(self) -> List[Union[torch.Tensor, nn.Module]]:
        raise NotImplementedError(
            f"Module [{type(self).__name__}] is missing the required "
            f'"iteration_parameters" function'
        )



"""Define a convolutional NSM module."""

import torch
from torch import nn

from typing import List, Sequence, Union, Optional

from .base import IterationLossModule


class MultiSimilarityMatching(IterationLossModule):
    """Multiple-target similarity matching circuit."""

    def __init__(
        self,
        encoders: Sequence[nn.Module],
        out_channels: int,
        tau: float = 0.1,
        max_iterations: int = 40,
        regularization: Union[str, Sequence[str]] = "weight",
        **kwargs,
    ):
        """Initialize the supervised similarity matching circuit.

        Some of the encoders can be skipped during the `forward()` call either by
        including fewer arguments than `len(encoders)` or by setting some to `None`.

        :param encoders: modules to use for encoding the inputs
        :param out_channels: number of output channels
        :param tau: factor by which to divide the competitor's learning rate
        :param max_iterations: maximum number of iterations to run in `forward()`
        :param regularization: type of encoder regularization to use; this can be a
            single string, or a sequence, to have different regularizations for each
            encoder; options are
            "weight":   use the encoders' parameters; regularization is added for all
                        the tensors returned by `encoder.parameters()`, as long as those
                        tensors are trainable (i.e., `requires_grad` is true)
            "whiten":   use a regularizer that encourages whitening XXX explain
            "none":     do not use regularization for the encoder; most useful to allow
                        for custom regularization, since lack of regularization leads to
                        unstable dynamics in many cases
        :param **kwargs: additional keyword arguments passed to `IterationLossModule`
        """
        super().__init__(max_iterations=max_iterations, **kwargs)

        self.encoders = nn.ModuleList(encoders)
        self.out_channels = out_channels
        self.tau = tau

        if isinstance(regularization, str):
            self.regularization = [regularization] * len(self.encoders)
        else:
            self.regularization = regularization

        for crt_reg in self.regularization:
            if crt_reg not in ["weight", "whiten", "none"]:
                raise ValueError(f"Unknown regularization {crt_reg}")

        self.competitor = nn.Linear(out_channels, out_channels, bias=False)
        torch.nn.init.eye_(self.competitor.weight)

        # make sure we maximize with respect to competitor weight...
        # ...and implement the learning rate ratio
        scaling = -1.0 / tau
        self.competitor.weight.register_hook(lambda g: g * scaling)

        self.y = torch.tensor([])

    def _encode(
        self, *args: Optional[torch.Tensor], keep_all: bool = False
    ) -> Sequence[torch.Tensor]:
        Wx = []
        for x, encoder in zip(args, self.encoders):
            if x is not None:
                Wx.append(encoder(x))
            elif keep_all:
                Wx.append(None)

        return Wx

    def pre_iteration(self, *args: Optional[torch.Tensor]):
        Wx = self._encode(*args)
        self._Wx = [_.detach() for _ in Wx]

        Wx_sum = self._Wx[0]
        for w in self._Wx[1:]:
            Wx_sum += w
        self._Wx_sum = Wx_sum

        self.y = torch.zeros_like(Wx[0])
        super().pre_iteration(*args)

    def iteration_set_gradients(self, *args: Optional[torch.Tensor]):
        with torch.no_grad():
            My = torch.einsum("ij,bj... -> bi...", self.competitor.weight, self.y)
            self.y.grad = My - self._Wx_sum

    def iteration_loss(self, *args: Optional[torch.Tensor]) -> torch.Tensor:
        """Loss function associated with the iteration.

        This is not actually used by the iteration, which instead uses manually
        calculated gradients (for efficiency).
        """
        assert self._Wx is not None
        loss = self._loss_no_reg(self._Wx, self.y, "sum")
        return loss / 4

    def post_iteration(self, *args: Optional[torch.Tensor]) -> torch.Tensor:
        super().post_iteration(*args)
        self._Wx = None
        self._Wx_sum = None

        return self.y.detach()

    def iteration_parameters(self) -> List[torch.Tensor]:
        return [self.y]

    def loss(self, *args: Optional[torch.Tensor]) -> torch.Tensor:
        y = args[-1]
        args = args[:-1]

        assert y is not None

        Wx = self._encode(*args)
        Wx_active = [_ for _ in Wx if _ is not None]
        loss = self._loss_no_reg(Wx_active, y, "mean")

        # competitor regularization
        M_reg = (self.competitor.weight**2).sum() / y.shape[1]
        loss -= M_reg

        # encoder regularization
        for encoder, regularization, crt_Wx in zip(
            self.encoders, self.regularization, Wx
        ):
            if regularization == "whiten":
                # this needs to match the scaling from _loss_no_reg!
                if crt_Wx is not None:
                    loss += 2 * (crt_Wx**2).mean()
            elif regularization == "weight":
                encoder_params = [_ for _ in encoder.parameters() if _.requires_grad]
                for weight in encoder_params:
                    loss += (weight**2).sum() * (2.0 / y.shape[1])
            elif regularization != "none":
                raise ValueError(f"Unknown regularization {regularization}")

        return loss

    def _loss_no_reg(
        self, Wx: Sequence[torch.Tensor], y: torch.Tensor, reduction: str
    ) -> torch.Tensor:
        """Compute the part of the loss without the regularization terms.

        :param Wx: encoded input, `self.encoder(x)`
        :param y: output (after iteration converges)
        :param reduction: "mean" or "sum"
        """
        My = torch.einsum("ij,bj... -> bi...", self.competitor.weight, y)
        yMy = (y * My).sum()

        loss = 2 * yMy

        if len(Wx) > 0:
            Wx_sum = Wx[0]
            for crt_Wx in Wx[1:]:
                Wx_sum += crt_Wx

            loss -= 4 * (y * Wx_sum).sum()

        if reduction == "mean":
            loss /= torch.numel(y)

        return loss


class SimilarityMatching(MultiSimilarityMatching):
    """Similarity matching circuit."""

    def __init__(self, encoder: nn.Module, out_channels: int, **kwargs):
        """Initialize the similarity matching circuit.

        This is a thin wrapper around `MultiSimilarityMatching` using a single target.

        :param encoder: module to use for encoding the inputs
        :param out_channels: number of output channels
        :param **kwargs: additional keyword arguments go to `MultiSimilarityMatching`
        """
        super().__init__(encoders=[encoder], out_channels=out_channels, **kwargs)

    @property
    def encoder(self) -> nn.Module:
        return self.encoders[0]


class SupervisedSimilarityMatching(MultiSimilarityMatching):
    """Supervised similarity matching circuit for classification."""

    def __init__(
        self,
        encoder: nn.Module,
        num_classes: int,
        out_channels: int,
        label_bias: bool = False,
        **kwargs,
    ):
        """Initialize the supervised similarity matching circuit.

        This is a wrapper that uses `MultiSimilarityMatching` with two encoders,
        `self.encoders == [encoder, label_encoder]`. The `label_encoder` is generated
        internally to map *floating-point* one-hot labels (which is what `forward()`
        expects) to a one-dimensional output of size `out_channels`. The `forward()`
        iteration is adapted to extend the output of the `label_encoder` to the same
        shape as the output from the `encoder()`, so it can be used in the similarity
        matching objective.

        Note that by default `"whiten"` regularization is used for the label encoder and
        `"weight"` regularization for the input encoder.

        :param encoder: module to use for encoding the inputs
        :param num_classes: number of classes for classification
        :param out_channels: number of output channels
        :param label_bias: set to true to include a bias term in the label encoder
        :param **kwargs: additional keyword arguments go to `MultiSimilarityMatching`
        """
        label_encoder = nn.Linear(num_classes, out_channels, bias=label_bias)
        if "regularization" not in kwargs:
            kwargs["regularization"] = ("weight", "whiten")
        super().__init__(
            out_channels=out_channels, encoders=[encoder, label_encoder], **kwargs
        )

    def _encode(
        self,
        x: torch.Tensor,
        label: Optional[torch.Tensor] = None,
        keep_all: bool = False,
    ) -> Sequence[Optional[torch.Tensor]]:
        # Wx will be shape [B, M, C1, ..., Ck], where B is batch size, M is out_channels
        Wx = self.encoders[0](x)
        if label is not None:
            # Wlabel0 will be shaoe [B, M]
            Wlabel0 = self.encoders[1](label)
            # we want to extend this to [B, M, C1, ..., Ck]
            extra_dim = Wx.ndim - 2
            Wlabel_unsqueezed = Wlabel0.reshape(Wlabel0.shape + (1,) * extra_dim)
            Wlabel = Wlabel_unsqueezed.expand(Wx.shape)

            return [Wx, Wlabel]
        else:
            if keep_all:
                return [Wx, None]
            else:
                return [Wx]



"""Define functions to validate circuit output."""

import torch
import torch.utils
import torch.utils.data
from torch import nn

from typing import Optional, Callable
from types import SimpleNamespace


def extract_embeddings(
    model: nn.Module,
    loader: torch.utils.data.DataLoader,
    progress: Optional[Callable] = None,
) -> SimpleNamespace:
    """Extract model embeddings and corresponding labels.

    This assumes a supervised-learning dataset, with samples of the form `(x, label)`.
    Note that the label is never actually used by the model.

    :param model: module to generate embeddings
    :param loader: iterable returning `(input, label)` pairs
    :param progress: progress indicator with `tqdm`-like interface
    :return: a namespace with members `output` and `label`, each a tensor containing the
        embeddings and the labels
    """
    was_training = model.training
    model.eval()

    # figure out device to send data to
    try:
        param = next(iter(model.parameters()))
        device = param.device
    except StopIteration:
        device = torch.device("cpu")

    # process the dataset
    results = SimpleNamespace(output=[], label=[])
    if progress is None:
        iterable = loader
    else:
        iterable = progress(loader)
    for x, label in iterable:
        x = x.to(device)

        y = model(x)
        results.output.append(y.detach().cpu())
        results.label.append(label.detach().cpu())

    # concatenate to tensors
    results.output = torch.cat(results.output)
    results.label = torch.cat(results.label)

    if was_training:
        model.train()

    return results



# _version.py generated by setuptools_scm
from ._version import __version__  # type: ignore

from .arch import IterationModule, IterationLossModule
from .arch import SimilarityMatching, MultiSimilarityMatching
from .arch import SupervisedSimilarityMatching
from .util import extract_embeddings

__all__ = [
    "IterationModule",
    "IterationLossModule",
    "SimilarityMatching",
    "MultiSimilarityMatching",
    "SupervisedSimilarityMatching",
    "extract_embeddings",
    "__version__",
]



from .base import IterationModule, IterationLossModule

from .similarity import SimilarityMatching, MultiSimilarityMatching
from .similarity import SupervisedSimilarityMatching

__all__ = [
    "IterationModule",
    "IterationLossModule",
    "SimilarityMatching",
    "MultiSimilarityMatching",
    "SupervisedSimilarityMatching",
]



"""Define a class to perform a ZCA transform."""

import numpy as np
import torch
from torch.utils.data import Dataset


def computeZCAMAtrix(X):
    # This function computes the ZCA matrix for a set of observables X where
    # rows are the observations and columns are the variables (M x C x W x H matrix)
    # C is number of color channels and W x H is width and height of each image

    X = np.asarray(X)

    # compute mean and std and normalize the data to -1 1 range with 1 std
    mean = X.mean(axis=0)
    std = X.std(axis=0)
    X_std = (X - mean) / std

    # reshape data from M x C x W x H to M x N where N=C x W x H
    X_std_flat = X_std.reshape(len(X_std), -1)

    # compute the covariance
    cov = np.cov(X_std_flat, rowvar=False)  # cov is (N, N)

    # singular value decomposition
    U, S, V = np.linalg.svd(cov)  # U is (N, N), S is (N,1) V is (N,N)
    # build the ZCA matrix which is (N,N)
    epsilon = 1e-5
    zca_matrix = np.dot(U, np.dot(np.diag(1.0 / np.sqrt(S + epsilon)), U.T))

    return (torch.from_numpy(zca_matrix).float(), mean, std)


class ZCATransformation(object):
    def __init__(self, transformation_matrix, transformation_mean):
        if transformation_matrix.size(0) != transformation_matrix.size(1):
            raise ValueError(
                "transformation_matrix should be square. Got "
                + "[{} x {}] rectangular matrix.".format(*transformation_matrix.size())
            )
        self.transformation_matrix = transformation_matrix
        self.transformation_mean = transformation_mean

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (N, C, H, W) to be whitened.
        Returns:
            Tensor: Transformed image.
        """
        count = tensor.size(1) * tensor.size(2) * tensor.size(3)
        if count != self.transformation_matrix.size(0):
            raise ValueError(
                "tensor and transformation matrix have incompatible shape."
                + "[{} x {} x {}] != ".format(*tensor[0].size())
                + "{}".format(self.transformation_matrix.size(0))
            )
        batch = tensor.size(0)

        flat_tensor = tensor.view(batch, -1)
        transformed_tensor = torch.mm(
            flat_tensor - self.transformation_mean, self.transformation_matrix
        )

        tensor = transformed_tensor.view(tensor.size())
        return tensor

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        format_string += str(self.transformation_matrix.numpy().tolist()) + ")"
        return format_string


class CustomImageDataset(Dataset):
    def __init__(self, X, Y, transform=None, target_transform=None):
        self.img_labels = Y
        self.img = X

    def __len__(self):
        return len(self.img_labels)

    def __getitem__(self, idx):
        return self.img[idx], self.img_labels[idx]



import pytest

import torch
from torch import nn

from typing import List

from unittest.mock import Mock

from pynsm.arch import IterationModule, IterationLossModule


class MockModule(IterationModule):
    def __init__(self, starts_converged: bool, max_iterations: int = 1000):
        super().__init__(max_iterations=max_iterations)
        self.is_converged = starts_converged
        self.n_calls = 0

    def iteration(self, *args, **kwargs):
        self.n_calls += 1

    def converged(self) -> bool:
        return self.is_converged


class MockModuleWithPrePost(IterationModule):
    def __init__(self):
        super().__init__(max_iterations=5)
        self.n_calls_pre = 0
        self.n_calls_post = 0

        self.last_call = None
        self.last_pre_call = None
        self.last_post_call = None
        self.last_conv_call = None

    def iteration(self, *args, **kwargs):
        self.last_call = (args, kwargs)

    def pre_iteration(self, *args, **kwargs):
        self.n_calls_pre += 1
        self.last_pre_call = (args, kwargs)

    def post_iteration(self, *args, **kwargs) -> str:
        self.n_calls_post += 1
        self.last_post_call = (args, kwargs)
        return "foobar"

    def converged(self, *args, **kwargs) -> bool:
        self.last_conv_call = (args, kwargs)
        return False


class MockLossModule(IterationLossModule):
    def __init__(self, n: int = 5, **kwargs):
        super().__init__(iteration_lr=0.1, **kwargs)

        torch.manual_seed(42)
        self.n = n
        self.decoy = nn.Linear(2, 3, bias=False)
        self.register_buffer("state", torch.randn(self.n))
        self.loss_history = []

    def iteration_loss(self, *args, **kwargs) -> torch.Tensor:
        loss = ((self.state + torch.ones(self.n)) ** 2).sum()
        self.loss_history.append(loss.item())
        return loss

    def iteration_parameters(self) -> List[torch.Tensor]:
        return [self.state]  # type: ignore

    def post_iteration(self, *args, **kwargs) -> str:
        super().post_iteration(*args, **kwargs)
        return "foobar"


def test_base_inherits_from_module():
    module = IterationModule()
    assert isinstance(module, nn.Module)


def test_forward_on_base_raises_not_implemented():
    module = IterationModule()
    with pytest.raises(NotImplementedError):
        module()


def test_iteration_called_until_max_iterations():
    n = 35
    module = MockModule(False, max_iterations=n)

    assert module.n_calls == 0
    module()
    assert module.n_calls == n


def test_iteration_ends_when_converged_is_true():
    module = MockModule(True)

    assert module.n_calls == 0
    module()
    assert module.n_calls == 1


def test_pre_and_post_iteration_are_called_once():
    module = MockModuleWithPrePost()

    assert module.n_calls_pre == 0
    assert module.n_calls_post == 0
    module()
    assert module.n_calls_pre == 1
    assert module.n_calls_post == 1


def test_args_kwargs_passed_to_iteration_and_pre_post_converged():
    module = MockModuleWithPrePost()
    module(2, 3, foo="bar")

    for last in [
        module.last_call,
        module.last_pre_call,
        module.last_post_call,
        module.last_conv_call,
    ]:
        assert last == ((2, 3), {"foo": "bar"})


def test_forward_returns_output_from_post_iteration():
    module = MockModuleWithPrePost()
    ret = module(2)

    assert ret == "foobar"


def test_pre_hook_called_before_pre_iteration():
    module = MockModuleWithPrePost()
    hook_called = [False]

    def hook(m):
        hook_called[0] = True
        assert m.last_pre_call is None
        assert m.last_call is None

    module.register_iteration_hook("pre", hook)
    module()

    assert hook_called[0]


def test_post_hook_called_after_post_iteration():
    module = MockModuleWithPrePost()
    hook_called = [False]

    def hook(m):
        hook_called[0] = True
        assert m.last_post_call is not None

    module.register_iteration_hook("post", hook)
    module()

    assert hook_called[0]


def test_iteration_hook_called_after_every_iteration():
    module = MockModuleWithPrePost()
    n_calls = [0]

    def hook(m):
        n_calls[0] += 1
        assert m.last_call is not None

    module.register_iteration_hook("iteration", hook)
    module()

    assert n_calls[0] == module.max_iterations


def test_register_multiple_hooks():
    module = MockModuleWithPrePost()
    hooks_called = [False, False]

    def hook1(m):
        hooks_called[0] = True

    def hook2(m):
        hooks_called[1] = True

    module.register_iteration_hook("iteration", hook1)
    module.register_iteration_hook("iteration", hook2)
    module()

    assert all(hooks_called)


def test_iteration_stops_if_iteration_hook_returns_truthful():
    module = MockModuleWithPrePost()
    n_calls = [0]

    def hook(m):
        n_calls[0] += 1
        return True

    module.register_iteration_hook("iteration", hook)
    module()

    assert n_calls[0] == 1


def test_iteration_idx_updated():
    n = 13
    module = MockModule(False, max_iterations=n)

    idxs = []
    module.register_iteration_hook("iteration", lambda m: idxs.append(m.iteration_idx))
    module()

    assert idxs == list(range(n))


def test_loss_of_base_loss_model_raises_not_implemented():
    module = IterationLossModule()
    with pytest.raises(NotImplementedError):
        module.iteration_loss()


def test_loss_model_forward_lowers_iteration_loss():
    module = MockLossModule()

    loss0 = module.iteration_loss().item()
    module()
    loss1 = module.iteration_loss().item()

    assert loss1 < loss0


def test_loss_model_forward_resets_iter_params_grad_state_to_false():
    module = MockLossModule()
    module()
    assert not module.state.requires_grad


def test_loss_model_change_optimizer():
    mock_optim = Mock()
    mock_optim_class = Mock(return_value=mock_optim)
    module = MockLossModule(iteration_optimizer=mock_optim_class)
    module()

    mock_optim_class.assert_called_once()
    mock_optim.zero_grad.assert_called()
    mock_optim.step.assert_called()


def test_loss_model_use_scheduler():
    mock_sched = Mock()
    mock_sched_class = Mock(return_value=mock_sched)
    module = MockLossModule(iteration_scheduler=mock_sched_class)
    module()

    mock_sched_class.assert_called_once()
    mock_sched.step.assert_called()


def test_loss_model_it_optim_kwargs_passed_to_optimizer():
    mock_optim = Mock()
    module = MockLossModule(
        iteration_optimizer=mock_optim, it_optim_kwargs={"foo": "bar"}
    )
    module()

    assert "foo" in mock_optim.call_args.kwargs
    assert mock_optim.call_args.kwargs["foo"] == "bar"


def test_loss_model_it_sched_kwargs_passed_to_scheduler():
    mock_sched = Mock()
    module = MockLossModule(
        iteration_scheduler=mock_sched, it_sched_kwargs={"foo": "bar"}
    )
    module()

    assert "foo" in mock_sched.call_args.kwargs
    assert mock_sched.call_args.kwargs["foo"] == "bar"


def test_loss_model_projection():
    # make sure we would have some negative elements without a projection
    module1 = MockLossModule(n=3)
    module1()
    assert torch.any(module1.state < 0)  # type: ignore

    module2 = MockLossModule(n=3, iteration_projection=nn.functional.relu)
    module2()
    assert torch.all(module2.state >= 0)  # type: ignore


def test_loss_model_projection_after_every_iteration():
    # make sure we would have some negative elements without a projection
    module1 = MockLossModule(n=3)
    module1()
    assert torch.any(module1.state < 0)  # type: ignore

    module2 = MockLossModule(n=3, iteration_projection=nn.functional.relu)
    module2.pre_iteration()
    module2.iteration()
    assert torch.all(module2.state >= 0)  # type: ignore


def test_pre_post_iteration_toggles_requires_grad_for_it_params():
    module = MockLossModule()
    module.pre_iteration()
    assert module.state.requires_grad

    module.post_iteration()
    assert not module.state.requires_grad


def test_pre_post_iteration_toggles_requires_grad_for_params():
    module = MockLossModule()
    params = list(module.parameters())
    assert len(params) > 0
    assert params[0].requires_grad

    module.pre_iteration()
    assert not params[0].requires_grad

    module.post_iteration()
    assert params[0].requires_grad


def test_pre_post_iteration_resets_requires_grad_for_params_to_what_it_was():
    module = MockLossModule()
    params = list(module.parameters())
    assert len(params) > 0
    params[0].requires_grad_(False)

    module.pre_iteration()
    assert not params[0].requires_grad

    module.post_iteration()
    assert not params[0].requires_grad


def test_loss_model_forward_returns_output_from_post_iteration():
    module = MockLossModule()
    ret = module(2)

    assert ret == "foobar"


def test_loss_model_updates_iteration_loss():
    module = MockLossModule()
    loss_history = []

    module.register_iteration_hook(
        "iteration", lambda m: loss_history.append(m.last_iteration_loss)
    )
    module()

    assert pytest.approx(loss_history) == module.loss_history


def test_override_backward_with_iteration_set_gradients():
    class MockLossModuleSetGrads(IterationLossModule):
        def __init__(self):
            super().__init__(iteration_lr=0.1)
            self.y = torch.zeros(5)

        def iteration_set_gradients(self, *args, **kwargs):
            self.y.grad = 1.0 * torch.arange(5)

        def iteration_parameters(self) -> List[torch.Tensor]:
            return [self.y]  # type: ignore

    module = MockLossModuleSetGrads()
    module()

    assert torch.allclose(module.y.grad, 1.0 * torch.arange(5))  # type: ignore



import pytest
from unittest.mock import Mock

import torch
from torch import nn

from typing import List, Tuple

from pynsm.util import extract_embeddings


@pytest.fixture
def linear_model() -> nn.Module:
    torch.manual_seed(21)
    return nn.Linear(5, 3)


@pytest.fixture
def data_loader() -> List[Tuple[torch.Tensor, torch.Tensor]]:
    torch.manual_seed(42)

    n = 12
    m = 5
    data = torch.randn(n, 5, m)
    labels = torch.randint(0, 8, (n, 5))

    loader = list(zip(data, labels))
    return loader


def test_extracts_labels_and_model_outputs(linear_model, data_loader):
    results = extract_embeddings(linear_model, data_loader)

    assert hasattr(results, "output")
    assert hasattr(results, "label")

    i = 0
    for x, label in data_loader:
        for bi in range(len(x)):
            assert torch.allclose(results.output[i], linear_model(x[bi]))
            assert results.label[i] == label[bi]
            i += 1


def test_sets_model_in_eval_mode(data_loader):
    model = Mock(return_value=torch.tensor([0.0, 0.0, 0.0, 0.0, 0.0]))
    model.parameters.return_value = [torch.tensor([])]
    extract_embeddings(model, data_loader)

    model.eval.assert_called_once()


def test_restores_model_to_train_mode(linear_model, data_loader):
    extract_embeddings(linear_model, data_loader)
    assert linear_model.training


def test_leaves_model_in_eval_mode_if_it_was_there(linear_model, data_loader):
    linear_model.eval()
    extract_embeddings(linear_model, data_loader)
    assert not linear_model.training


def test_progress_called_with_loader(linear_model, data_loader):
    progress = Mock(side_effect=lambda x: x)
    extract_embeddings(linear_model, data_loader, progress=progress)
    progress.assert_called_once()
